/*
 * (c) Copyright Robin Hill 2004
 *
 * For license terms, see the file COPYING in this directory.
 *
 * Wrapper for wxFFile using customised Tell(), Seek() and Length()
 * functions changed for 64-bit support, and with added Resize() function.
 *
 */

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
	#pragma hdrstop
#endif

#ifndef WX_PRECOMP
	#include <wx/wx.h>
#endif

#ifdef __WINDOWS__
	#include <windows.h>
	#include <io.h>
#endif
#ifdef __UNIX__
	#include <unistd.h>
#endif

#include "frFFile.h"

bool frFFile::Seek(fr_off_t ofs, wxSeekMode mode)
{
	int origin;

	wxLogTrace(wxT("frTRACE_functions"), wxT("Seek: Entry"));
	wxASSERT(IsOpened());

	switch ( mode ) {
		default:
			// fall through

		case wxFromStart:
			origin = SEEK_SET;
			break;

		case wxFromCurrent:
			origin = SEEK_CUR;
			break;

		case wxFromEnd:
			origin = SEEK_END;
			break;
	}

#ifdef __UNIX__
	if ( fseeko(wxfile->fp(), ofs, origin) != 0 ) {
#else
#ifdef __WINDOWS__
	if ( _lseeki64(fileno(wxfile->fp()), ofs, origin) == -1 ) {
#else
	if ( ofs > (fr_off_t)0x80000000) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Seek: Error"));
		return FALSE;
	}
	if ( fseek(wxfile->fp(), (long)ofs, origin) != 0 ) {
#endif /* __WINDOWS__ */
#endif /* __UNIX__ */
		wxLogTrace(wxT("frTRACE_functions"), wxT("Seek: Error"));
		return FALSE;
	}

	wxLogTrace(wxT("frTRACE_functions"), wxT("Seek: Exit"));
	return TRUE;
}

fr_off_t frFFile::Tell() const
{
	wxLogTrace(wxT("frTRACE_functions"), wxT("Tell: Entry"));

#ifdef __UNIX__
	fr_off_t  rc = ftello(wxfile->fp());
#else
#ifdef __WINDOWS__
	fr_off_t  rc = _telli64(fileno(wxfile->fp()));
#else
	fr_off_t  rc = ftell(wxfile->fp());
#endif /* __WINDOWS__ */
#endif /* __UNIX__ */
	if ( rc == -1 ) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Tell: Error"));
		return rc;
	}

	wxLogTrace(wxT("frTRACE_functions"), wxT("Tell: Exit"));
	return rc;
}

fr_off_t frFFile::Length() const
{
	wxLogTrace(wxT("frTRACE_functions"), wxT("Length: Entry"));

	frFFile& self = *(frFFile *)this;   // const_cast

	fr_off_t posOld = Tell();
	if ( posOld != (fr_off_t)-1 ) {
		if ( self.SeekEnd() ) {
			fr_off_t len = Tell();

			(void)self.Seek(posOld);

			wxLogTrace(wxT("frTRACE_functions"), wxT("Length: Exit"));
			return len;
		}
	}

	wxLogTrace(wxT("frTRACE_functions"), wxT("Length: Error"));
	return (fr_off_t)-1;
}

bool frFFile::Resize(const fr_off_t size)
{
	wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Entry"));

#ifndef __UNIX__
#ifndef __WINDOWS__
	fr_off_t        oldsize,
	                difference;
	unsigned char   *buffer;
	wxInt32         write,
	                wrote;
	const wxInt32   bufsize = 1048576;
#endif /* not __WINDOWS__ */
	frFFile&        self = *(frFFile *)this;   // const_cast
#endif /* not __UNIX__ */

#ifdef __UNIX__
	if (ftruncate(fileno(wxfile->fp()), size) < 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
		return FALSE;
	}
#else /* not __UNIX__ */
#ifdef __WINDOWS__
	if (! self.Seek(size, wxFromStart)) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
		return FALSE;
	}
	if (SetEndOfFile((HANDLE)_get_osfhandle(fileno(wxfile->fp()))) == 0) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
		return FALSE;
	}
#else   
	oldsize = self.Length();

	if (oldsize < size) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
		return FALSE;
	} else {
		difference = size - oldsize;
		wxLogTrace(wxTRACE_MemAlloc, wxT("Resize: allocating buffer"));
		buffer = new unsigned char[bufsize];
		memset(buffer, 0, bufsize);
		if (! self.SeekEnd()) {
			wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
			return FALSE;
		}
		while (difference > 0) {
			if (difference > bufsize)
				write = bufsize;
			else
				write = difference;
			wrote = self.Write(buffer, write);
			if (wrote < 0) {
				wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
				wxLogTrace(wxTRACE_MemAlloc, wxT("Resize: freeing buffer"));
				delete[] buffer;
				return FALSE;
			}
			difference -= wrote;
		}
		infile->Flush();
		wxLogTrace(wxTRACE_MemAlloc, wxT("Resize: freeing buffer"));
		delete[] buffer;
	}
#endif /* __WINDOWS__ */
	if (! self.Seek(0, wxFromStart)) {
		wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Error"));
		return FALSE;
	}
#endif /* __UNIX__ */
	wxLogTrace(wxT("frTRACE_functions"), wxT("Resize: Exit"));
	return TRUE;
}
