/*
 * (c) Copyright Robin Hill 2002
 *
 * For license terms, see the file COPYING in this directory.
 */

#include <stdio.h>
#include <curses.h>
#include <wx/app.h>
#include <wx/string.h>
#include <wx/log.h>
#include <wx/cmdline.h>
#include <wx/socket.h>
#include <wx/longlong.h>
#include <wx/utils.h>
#include "frFFile.h"
#include "fr_backend.h"
#include "defines.h"
#include "fr_cli.h"

int main(int argc, char **argv)
{
// Try to initialize the wxWindows system.  This is needed for anything else
// to work!
	if ( !wxInitialize() )
	{
		fprintf(stderr, "Failed to initialize the wxWindows library, aborting.");
		exit(1);
	}

	wxString        header;

// Debug mode settings.
#ifdef __WXDEBUG__
// Output debug messages to the frcli_errors.txt file
	FILE *errors;
	wxLog *console;
	errors = fopen("frcli_errors.txt", "w");
	wxLog *logger = new wxLogStderr(errors);
	console = wxLog::SetActiveTarget(logger);
// Add an extra trace option for function entry/exit
//	wxLog::AddTraceMask(wxT("frTRACE_functions"));
//	wxLog::AddTraceMask(wxT("frTRACE_net"));
//	wxLog::AddTraceMask(wxTRACE_MemAlloc);
//	wxLog::AddTraceMask(wxT("frTRACE_debug"));
#endif

// Set/clear the global parameters
	blocksize = DEF_BLOCKSIZE / 1024;
	port = DEF_PORTNUM;
	passive = FALSE;
	ignore = FALSE;
	quiet = FALSE;
	ChkFile.Clear();
	InputFile.Clear();
	ServerAddress.Clear();

// Create a header giving some program info
	header.Printf(wxT("%s version %s, built %s (%s)\n"), wxT(PROGRAM), wxT(VERSION),  wxT(__DATE__), wxT(__TIME__));

// Setup the command line options
	static const wxCmdLineEntryDesc cmdLineDesc[] =
	{
		{ wxCMD_LINE_SWITCH, wxT("?"), wxT("help"),       wxT("Show usage\n"), wxCMD_LINE_VAL_NONE, wxCMD_LINE_OPTION_HELP },

		{ wxCMD_LINE_SWITCH, wxT("s"), wxT("server"),     wxT("Server mode") },
		{ wxCMD_LINE_SWITCH, wxT("c"), wxT("client"),     wxT("Client mode") },
		{ wxCMD_LINE_SWITCH, wxT("r"), wxT("crc"),	wxT("Generate or check CRC list") },
		{ wxCMD_LINE_SWITCH, wxT("h"), wxT("hash"),       wxT("Generate md5sum\n") },
		{ wxCMD_LINE_OPTION, wxT("t"), wxT("type"),       wxT("Hash type\n") },

		{ wxCMD_LINE_OPTION, wxT("n"), wxT("servername"), wxT("Server to connect to"), wxCMD_LINE_VAL_STRING },
		{ wxCMD_LINE_OPTION, wxT("i"), wxT("input"),      wxT("CRC input file"), wxCMD_LINE_VAL_STRING },
		{ wxCMD_LINE_OPTION, wxT("o"), wxT("output"),     wxT("CRC output file"), wxCMD_LINE_VAL_STRING },
		{ wxCMD_LINE_OPTION, wxT("p"), wxT("port"),       wxT("Server port to use"), wxCMD_LINE_VAL_NUMBER },
		{ wxCMD_LINE_OPTION, wxT("b"), wxT("blocksize"),  wxT("Blocksize to use (kb)\n"), wxCMD_LINE_VAL_NUMBER },

		{ wxCMD_LINE_SWITCH, wxT("a"), wxT("passive"),    wxT("Use passive mode") },
		{ wxCMD_LINE_SWITCH, wxT("f"), wxT("ignore_filesize"),    wxT("Ignore difference in filesizes\n") },
		{ wxCMD_LINE_SWITCH, wxT("q"), wxT("quiet"),      wxT("Don't output progress information\n") },

		{ wxCMD_LINE_PARAM,  NULL, NULL, wxT("input file"), wxCMD_LINE_VAL_STRING },

		{ wxCMD_LINE_NONE }
	};

	wxCmdLineParser parser(cmdLineDesc, argc, argv);
	parser.SetLogo(header);

// Parse the command line
	wxLogDebug(wxT("Checking command-line parameters"));

	switch (parser.Parse()) {
// If it's a help request then give some extra info
		case -1:
			header.Printf(wxT("(c) COPYRIGHT %s %s - %s\n"), wxT(CONCEPTDATE), wxT(AUTHOR), wxT(XWARE));
			header.Append(wxString::Format(wxT("EMAIL: %s"), wxT(EMAIL)));
			wxLogMessage(header);
			wxUninitialize();
			exit(0);
		case 0:
// If it parsed okay then continue
			break;
		default:
// If it failed to parse then just exit
			wxUninitialize();
			exit(1);
	}

// Check whether the necessary combinations of options have been given
	wxLogDebug(wxT("Checking for illegal combinations or missing parameters"));
	if ((parser.Found(wxT("i"), &ChkFile) && parser.Found(wxT("o"), &ChkFile)) ||
	    (parser.Found(wxT("c")) && (! parser.Found(wxT("n"), &ServerAddress))) ||
	    (parser.Found(wxT("h")) && (! parser.Found(wxT("t"), &HashType))) ||
	    (parser.Found(wxT("r")) && (! (parser.Found(wxT("i"), &ChkFile) || parser.Found(wxT("o"), &ChkFile))))) {
		parser.Usage();
		wxUninitialize();
		exit(1);
	}

// Read in the parameters set
	wxLogDebug(wxT("Checking for port"));
	parser.Found(wxT("p"), &port);
	wxLogDebug(wxT("Checking for blocksize"));
	parser.Found(wxT("b"), &blocksize);
	wxLogDebug(wxT("Getting input file"));
	InputFile = parser.GetParam(0);
	wxLogDebug(wxT("Checking for passive mode"));
	if (parser.Found(wxT("a")))
		passive = TRUE;
	wxLogDebug(wxT("Checking whether ignoring filesize differences"));
	if (parser.Found(wxT("f")))
		ignore = TRUE;
	wxLogDebug(wxT("Checking whether to skip outputting status info"));
	if (parser.Found(wxT("q")))
		quiet = TRUE;

// Check which mode was requested
	wxLogDebug(wxT("Checking which mode to run in"));
	if (parser.Found(wxT("c"))) RunJob(header, MODE_CLIENT_START);
	else if (parser.Found(wxT("s"))) RunJob(header, MODE_SERVER_START);
	else if (parser.Found(wxT("r")) && parser.Found(wxT("o"), &ChkFile)) RunJob(header, MODE_CRC_CREATE);
	else if (parser.Found(wxT("r")) && parser.Found(wxT("i"), &ChkFile)) RunJob(header, MODE_CRC_VERIFY);
	else if (parser.Found(wxT("h"))) {
		wxLogDebug(wxT("Checking which hash type to use"));
		if (! HashType.CmpNoCase(wxT("md5"))) RunJob(header, MODE_FH_MD5);
		else if (! HashType.CmpNoCase(wxT("edonkey"))) RunJob(header, MODE_FH_ED2K);
		else if (! HashType.CmpNoCase(wxT("ed2k"))) RunJob(header, MODE_FH_ED2K);
		else if (! HashType.CmpNoCase(wxT("crc32"))) RunJob(header, MODE_FH_CRC);
		else if (! HashType.CmpNoCase(wxT("tth"))) RunJob(header, MODE_FH_TTH);
		else if (! HashType.CmpNoCase(wxT("crc"))) RunJob(header, MODE_FH_CRC);
		else {
			header.Append(wxT("Unrecognised hash type - known types are:\ned2k, md5, crc32\n"));
			wxLogMessage(header);
			wxUninitialize();
			exit(1);
		}
	} else {
		header.Append(wxT("No operation requested - exiting\n"));
		wxLogMessage(header);
	}


	wxUninitialize();
    	return 0;
}

//
// RunJob
//
// This is passed a header string and the job mode as inputs.  It then
// creates the curses interface and runs the relevant task from the backend.
void RunJob(wxString header, wxUint32 mode)
{
	FRCLI_BackEnd   backend;
	wxUint32        bufsize = blocksize * 1024;
	WINDOW          *screen;

	if (! quiet) {
// Initialise the curses interface
		screen = initscr();
		leaveok(screen, TRUE);
		intrflush(screen, FALSE);

// Print out the header info
		mvaddnstr(ROW_HEAD, 0, header.mb_str(), COLS);
// Draw the progress bar
		mvaddch(ROW_PROG1, 0, ACS_ULCORNER);
		hline(0, COLS - 2);
		mvaddch(ROW_PROG1, COLS -1, ACS_URCORNER);
		mvaddch(ROW_PROG2, 0, ACS_VLINE);
		mvaddch(ROW_PROG2, COLS - 1, ACS_VLINE);
		mvaddch(ROW_PROG3, 0, ACS_LLCORNER);
		hline(0, COLS - 2);
		mvaddch(ROW_PROG3, COLS - 1, ACS_LRCORNER);
		refresh();
	}
// Run the relevant backend job
	switch (mode) {
		case MODE_FH_ED2K:
			backend.CreateFileHashED2k(InputFile, DEF_BLOCKSIZE);
			break;
		case MODE_FH_MD5:
			backend.CreateFileHashMD5(InputFile, DEF_BLOCKSIZE);
			break;
		case MODE_FH_TTH:
			backend.CreateFileHashTTH(InputFile, DEF_BLOCKSIZE);
			break;
		case MODE_FH_CRC:
			backend.CreateFileHashCRC(InputFile, DEF_BLOCKSIZE);
			break;
		case MODE_CRC_CREATE:
			backend.CreateCRC(InputFile, ChkFile, bufsize);
			break;
		case MODE_CRC_VERIFY:
			backend.VerifyCRC(InputFile, ChkFile);
			break;
		case MODE_CLIENT_START:
			backend.Client(InputFile, ServerAddress, (unsigned short)port, bufsize, passive);
			break;
		case MODE_SERVER_START:
			backend.Server(InputFile, (unsigned short)port, passive);
			break;
	}
	endwin();
	printf("%s\n", (const char *)wxConvCurrent->cWX2MB(backend.lastMessage));
}

FRCLI_BackEnd::FRCLI_BackEnd()
{
}

void FRCLI_BackEnd::SendError(const wxString message)
{
	lastMessage = message;
	if (! quiet) {
		move(ROW_RES2, 0);
		clrtoeol();
		mvaddnstr(ROW_RES1, 0, message.mb_str(), COLS * 2);
		clrtoeol();
		refresh();
	}
}

void FRCLI_BackEnd::SendAlert(const wxString message)
{
	lastMessage = message;
	if (! quiet) {
		move(ROW_RES2, 0);
		clrtoeol();
		mvaddnstr(ROW_RES1, 0, message.mb_str(), COLS * 2);
		clrtoeol();
		refresh();
	}
}

void FRCLI_BackEnd::SendProgress(wxInt8 progress)
{
	if (progress > 100) progress = 100;
	if (! quiet) {
		if (progress == 0) {
			move(ROW_PROG2, 1);
			clrtoeol();
			mvaddch(ROW_PROG2, COLS - 1, ACS_VLINE);
			refresh();
		} else {
			mvhline(ROW_PROG2, 1, ACS_BLOCK, ((COLS - 2) * progress) / 100);
			refresh();
		}
	}
}

void FRCLI_BackEnd::SendStatus(const wxString message)
{
	lastMessage = message;

	if (! quiet) {
		move(ROW_STAT2, 0);
		clrtoeol();
		mvaddnstr(ROW_STAT1, 0, message.mb_str(), COLS * 2);
		clrtoeol();
		refresh();
	}
}

void FRCLI_BackEnd::SendServerInfo(const wxString message)
{
	if (! quiet) {
		mvaddnstr(ROW_SCINF, 0, message.mb_str(), COLS);
		clrtoeol();
		refresh();
	}
}

void FRCLI_BackEnd::SendHash(const wxString message)
{
	lastMessage = message;
	if (! quiet) {
		move(ROW_RES2, 0);
		clrtoeol();
		mvaddnstr(ROW_RES1, 0, message.mb_str(), COLS * 2);
		clrtoeol();
		refresh();
	}
}

bool FRCLI_BackEnd::CheckExit()
{
	return FALSE;
}

bool FRCLI_BackEnd::ContinueWrongFilesize(const wxString message)
{
	if (! ignore)
		SendError(message);
	return ignore;
}

void FRCLI_BackEnd::Sleep(unsigned long duration)
{
	wxMilliSleep(duration);
}
